﻿using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Xml;
using Microsoft.Xrm.Client;
using Microsoft.Xrm.Client.Services;
using Microsoft.Xrm.Sdk.Messages;
using Microsoft.Xrm.Sdk.Query;

namespace AppFx.DownloadStream.CRM
{   
    public class CrmDownloader : IDisposable, IDownload
    {
        private readonly OrganizationService _crmService;
        private readonly string _fetchXmlRequest;        
        private readonly string _pagingCookie = null;
        private int _pageNumber = 1;
        private bool _firstRead = true;
        private bool _lastRead = false;

        public bool MorePages { get; set; }
        public string MessageStart { get; set; }
        public string MessageEnd { get; set; }        
        public int FetchCount { get; set; }
        //private int RecordCount { get; set; }

        public CrmDownloader(string crmConnectionString, string fetchXmlRequest)
        {
            var crmConnection = CrmConnection.Parse(crmConnectionString);
            _crmService = new OrganizationService(crmConnection);
            _fetchXmlRequest = fetchXmlRequest;            
            MorePages = true;
            FetchCount = 50;
        }

        public CrmDownloader(OrganizationService crmService, string fetchXmlRequest)
        {
            _crmService = crmService;
            _fetchXmlRequest = fetchXmlRequest;            
            MorePages = true;
            FetchCount = 50;
        }

        public byte[] GetNext()
        {
            byte[] nextBytes = null;
            if (_firstRead)
            {
                var messageStart = GetMesageStart();
                var firstPage = GetNextPageFromCrm();
                nextBytes = new byte[messageStart.Length + firstPage.Length];
                messageStart.CopyTo(nextBytes, 0);
                firstPage.CopyTo(nextBytes, messageStart.Length);
                _firstRead = false;
                return nextBytes;
            }

            if (_firstRead == false)
            {
                if (_lastRead)
                {
                    nextBytes = new byte[0];
                    return nextBytes;
                }
                else
                {
                    if (MorePages)
                    {
                        nextBytes = GetNextPageFromCrm();                        
                        if (MorePages == false)
                        {
                            var endMessageBytes = GetMesageEnd();
                            var mergedNextBytes = new byte[nextBytes.Length + endMessageBytes.Length];
                            nextBytes.CopyTo(mergedNextBytes, 0);
                            endMessageBytes.CopyTo(mergedNextBytes, nextBytes.Length);
                            _lastRead = true;
                            return mergedNextBytes;
                        }
                    }
                    else
                    {
                        nextBytes = GetMesageEnd();
                        _lastRead = true;
                    }
                }
            }
            return nextBytes;
        }


        private byte[] GetNextPageFromCrm()
        {
            // Build fetchXml string with the placeholders.
            var xml = CreateXml(_fetchXmlRequest, _pagingCookie, _pageNumber, FetchCount);

            // Excute the fetch query and get the xml result.
            var fetchRequest1 = new RetrieveMultipleRequest
            {
                Query = new FetchExpression(xml)
            };

            var returnCollection =
                ((RetrieveMultipleResponse)_crmService.Execute(fetchRequest1)).EntityCollection;

            Trace.WriteLine("Retrieving records from CRM");
            MemoryStream dataStream = new MemoryStream();
            foreach (var crmEntity in returnCollection.Entities)
            {                
                var formatter = new FlattenedXmlFormatter();
                formatter.PrefixRootEntityAttributes = true;
                formatter.Format(dataStream, crmEntity);
            }

            // Check for morerecords, if it returns 1.
            if (returnCollection.MoreRecords)
            {
                // Increment the page number to retrieve the next page.
                _pageNumber++;
            }
            else
                MorePages = false;

            return dataStream.ToArray();
        }

        private byte[] GetMesageStart()
        {
            using (var ms = new MemoryStream())
            {
                var writer = new StreamWriter(ms);
                writer.Write(MessageStart);
                writer.Flush();
                ms.Flush();
                return ms.ToArray();
            }
        }

        private byte[] GetMesageEnd()
        {
            using (var ms = new MemoryStream())
            {
                var writer = new StreamWriter(ms);
                writer.Write(MessageEnd);
                writer.Flush();
                ms.Flush();
                return ms.ToArray();
            }
        }

        public void Dispose()
        {
            if (_crmService != null)
            {
                _crmService.Dispose();
            }
        }

        private string CreateXml(string xml, string cookie, int page, int count)
        {
            var stringReader = new StringReader(xml);
            var reader = new XmlTextReader(stringReader);

            // Load document
            var doc = new XmlDocument();
            doc.Load(reader);

            return CreateXml(doc, cookie, page, count);
        }

        private string CreateXml(XmlDocument doc, string cookie, int page, int count)
        {
            var attrs = doc.DocumentElement.Attributes;

            if (cookie != null)
            {
                var pagingAttr = doc.CreateAttribute("paging-cookie");
                pagingAttr.Value = cookie;
                attrs.Append(pagingAttr);
            }

            var pageAttr = doc.CreateAttribute("page");
            pageAttr.Value = Convert.ToString(page);
            attrs.Append(pageAttr);

            var countAttr = doc.CreateAttribute("count");
            countAttr.Value = Convert.ToString(count);
            attrs.Append(countAttr);

            var sb = new StringBuilder(1024);
            var stringWriter = new StringWriter(sb);

            var writer = new XmlTextWriter(stringWriter);
            doc.WriteTo(writer);
            writer.Close();

            return sb.ToString();
        }

    }
}
