﻿using System;
using System.Collections.Generic;
using System.IO;

namespace AppFx.DownloadStream
{
    public class GenericDownloadStream : Stream, IDisposable
    {
        private readonly Queue<byte> _writeQueue;         
        readonly IDownload _downloader;

        public GenericDownloadStream(IDownload downloader)
        {
            _downloader = downloader;
            _writeQueue = new Queue<byte>();
        }


        public override int Read(byte[] buffer, int offset, int count)
        {
            int bytesWritten = 0;//number of bytes written to output buffer; 
            var thisRead = new List<byte>();

            //first of all - use any 'left over bytes' from previous read (left over bytes will exist if buffer was smaller than any one line so we could not return the entire line to the caller)
            CopyFromWriteQueue(thisRead, count);
         
            while (_writeQueue.Count < count)
            {
                var nextPage = _downloader.GetNext();
                if (nextPage.Length == 0)
                {
                    break;
                }

                //Add data to write queue
                foreach(var nextByte in nextPage)
                    _writeQueue.Enqueue(nextByte);
            }

            //Get data from write queue
            CopyFromWriteQueue(thisRead, count);

            bytesWritten = thisRead.Count;            
            Buffer.BlockCopy(thisRead.ToArray(), 0, buffer, 0, thisRead.Count);
            return bytesWritten;
        }

        private void CopyFromWriteQueue(List<byte> thisRead, int count)
        {            
            var remaining = count - thisRead.Count;

            //The buffer is already full
            if (remaining == 0)
                return;

            //if we have no left over bytes
            if(_writeQueue.Count == 0)
                return;

            //if we have place for the entire 'left over'
            if (remaining >= _writeQueue.Count)
            {
                for (var index = 0; index < remaining; index++)
                {
                    if (_writeQueue.Count == 0)
                        break;

                    thisRead.Add(_writeQueue.Dequeue());
                }
            }

            //if we don't have place for the entire left over
            if (remaining < _writeQueue.Count)
            {
                for (var index = 0; index < count; index++)
                    thisRead.Add(_writeQueue.Dequeue());
            }
        }
        


        #region Stream overrides
        public override bool CanSeek
        {
            get { 
                //as the stream is bufferring, additional code would have been required to support seeking 
                //(mostly getting read of the buffer on seek), but this is not needed for the BizTalk scenario 
                //(BizTalk itself will never seek the stream), so I've avoided adding this code
                return false; }
        }
        public override long Position
        {
            get
            {
                return 0;                
            }
            set
            {
                throw new NotSupportedException("this is not seekable");
            }
        }
        public override bool CanWrite
        {
            get
            {
                return false; //stream is non-writable due to the way it buffers data, but this is not a requirement usually in this context
            }
        }
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotSupportedException("this is not seekable");
        }

        public override void SetLength(long value)
        {
            throw new NotSupportedException("this is not writable");
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotSupportedException("this is not writable");
        }

        #endregion

        #region Stream overrides - all of these are forwarded to the underlying stream
        public override bool CanRead
        {
            get { return true;// return _inputStream.CanRead; 
            }
        }

        public override void Flush()
        {
            
        }

        public override long Length
        {
            get { return 0;
            }
        }

        #endregion

        #region IDisposable Members

        void IDisposable.Dispose()
        {
            Close();
        }

        #endregion

    } 
}

